
/*
 * shvar.c
 *
 * Implementation of non-destructively reading/writing files containing
 * only shell variable declarations and full-line comments.
 *
 * Includes explicit inheritance mechanism intended for use with
 * Red Hat Linux ifcfg-* files.  There is no protection against
 * inheritance loops; they will generally cause stack overflows.
 * Furthermore, they are only intended for one level of inheritance;
 * the value setting algorithm assumes this.
 *
 * Copyright 1999,2000 Red Hat, Inc.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "shvar.h"

/* Open the file <name>, return shvarFile on success, NULL on failure */
shvarFile *
svNewFile(const char *name)
{
	shvarFile *s = calloc(sizeof(*s), 1);

	s->fd = open(name, O_RDONLY);
	s->fileName = strdup(name);

	if (s->fd != -1)
	{
		struct stat buf;
		char   *p, *q;

		if (fstat(s->fd, &buf) < 0)
			goto bail;
		s->arena = g_malloc0(buf.st_size + 1);

		if (read(s->fd, s->arena, buf.st_size) < 0)
			goto bail;

		/* we'd use g_strsplit() here, but we want a list, not an array */
		for (p = s->arena; (q = strchr(p, '\n')) != NULL; p = q + 1)
		{
			s->lineList =
				g_list_append(s->lineList,
					      g_strndup(p, q - p));
		}

		return s;
	}

      bail:
	if (s->fd != -1)
		close(s->fd);
	if (s->arena)
		g_free(s->arena);
	if (s->fileName)
		free(s->fileName);
	free(s);
	return NULL;
}

/* remove escaped characters in place */
static void
unescape(char *s)
{
	int     len, i;

	len = strlen(s);
	if ((s[0] == '"' || s[0] == '\'') && s[0] == s[len - 1])
	{
		i = len - 2;
		memmove(s, s + 1, i);
		s[i + 1] = '\0';
		len = i;
	}
	for (i = 0; i < len; i++)
	{
		if (s[i] == '\\')
		{
			memmove(s + i, s + i + 1, len - (i + 1));
			len--;
		}
		s[len] = '\0';
	}
}

/* Get the value associated with the key, and leave the current pointer
 * pointing at the line containing the value.  The char* returned MUST
 * be freed by the caller.
 */
char   *
svGetValue(shvarFile * s, const char *key)
{
	char   *value = NULL;
	char   *line;
	char   *keyString;
	int     len;

	g_assert(s);
	g_assert(key);

	keyString = g_malloc0(strlen(key) + 2);
	strcpy(keyString, key);
	keyString[strlen(key)] = '=';
	len = strlen(keyString);

	for (s->current = s->lineList; s->current;
	     s->current = s->current->next)
	{
		line = s->current->data;
		if (!strncmp(keyString, line, len))
		{
			value = g_strdup(line + len);
			unescape(value);
			break;
		}
	}
	g_free(keyString);

	if (value)
	{
		if (value[0])
		{
			return value;
		} else
		{
			g_free(value);
			return NULL;
		}
	}
	if (s->parent)
		value = svGetValue(s->parent, key);
	return value;
}

/* return 1 if <key> resolves to any truth value (e.g. "yes", "y", "true")
 * return 0 if <key> resolves to any non-truth value (e.g. "no", "n", "false")
 * return <default> otherwise
 */
int
svTrueValue(shvarFile * s, const char *key, int def)
{
	char   *tmp;
	int     returnValue = def;

	tmp = svGetValue(s, key);
	if (!tmp)
		return returnValue;

	if ((!strcasecmp("yes", tmp)) ||
	    (!strcasecmp("true", tmp)) ||
	    (!strcasecmp("t", tmp)) || (!strcasecmp("y", tmp)))
		returnValue = 1;
	else if ((!strcasecmp("no", tmp)) ||
		 (!strcasecmp("false", tmp)) ||
		 (!strcasecmp("f", tmp)) || (!strcasecmp("n", tmp)))
		returnValue = 0;

	g_free(tmp);
	return returnValue;
}

/* Close the file descriptor (if open) and delete the shvarFile.
 * Returns -1 on error and 0 on success.
 */
int
svCloseFile(shvarFile * s)
{

	g_assert(s);

	if (s->fd != -1)
		close(s->fd);

	g_free(s->arena);
	for (s->current = s->freeList; s->current;
	     s->current = s->current->next)
	{
		g_free(s->current->data);
	}
	g_free(s->fileName);
	g_list_free(s->freeList);
	g_list_free(s->lineList);	/* implicitly frees s->current */
	g_free(s);
	return 0;
}
